///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

part of haystack.hgrid;

/// HDict is an immutable map of name/HVal pairs.  Use HDictBuilder
/// to construct a HDict instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
abstract class HDict extends HVal with MapMixin<String, HVal?> {
  /// Singleton for empty set of tags.
  static final HDict EMPTY = new _MapImpl(
      new Map<String, HVal>()); //it was new MapImpl(new HashMap(11));

  factory HDict(Map<String, HVal?> map) = _MapImpl;

  HDict._();

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /// Returns if size is zero.
  @override
  bool get isEmpty => size == 0;

  /// Returns number of tag name/value pairs
  int get size;

  /// Returns if the given tag is present.
  ///
  /// Accordingly to Axon's dict interface returns null if key exists and has
  /// null value. Use [containsKey] to check key existence.
  bool has(String name) => get(name, checked: false) != null;

  bool hasTyped<T>(String name) {
    final val = get(name, checked: false);

    return val != null && val is T;
  }

  /// Returns if the given tag is not present.
  ///
  /// Same as [has] method, returns true if key's value is null. Read more in
  /// [has] docs.
  bool missing(String name) => get(name, checked: false) == null;

  /// Iterates keys and values in keys order and skipping null values.
  ///
  /// Again: IT SKIPS NULL VALUES.
  @override
  void forEach(void action(String key, HVal value)) {
    for (String key in keys) {
      if (this[key] != null) {
        action(key, this[key]!);
      }
    }
  }

  /// Get a tag by name.
  ///
  /// If not found and checked if false then
  /// Returns null, otherwise throw UnknownNameException.
  HVal? get(String name, {bool checked = true});

  /// Get the "id" tag as HRef.
  HRef? id() => getRef("id", checked: false);

  /// Get display string for this entity:
  ///   - dis tag
  ///   - id tag
  String dis() {
    HVal? v;
    v = get("dis", checked: false);
    if (v is HStr) return v.val;

    v = get("id", checked: false);
    if (v != null) return (v as HRef).getDis();

    return "????";
  }

//////////////////////////////////////////////////////////////////////////
// Get Conveniences
//////////////////////////////////////////////////////////////////////////

  /// Get tag as HBool or raise UnknownNameException or ClassCastException.
  bool? getBool(String name, {bool checked = true}) =>
      (get(name, checked: checked) as HBool?)?.val;

  /// Get tag as HStr or raise UnknownNameException or ClassCastException.
  String? getStr(String name, {bool checked = true}) =>
      (get(name, checked: checked) as HStr?)?.val;

  /// Get tag as HRef or raise UnknownNameException or ClassCastException.
  HRef? getRef(String name, {bool checked = true}) =>
      get(name, checked: checked) as HRef?;

  /// Get tag as HNum or raise UnknownNameException or ClassCastException.
  int? getInt(String name, {bool checked = true}) =>
      (get(name, checked: checked) as HNum?)?.val.toInt();

  /// Get tag as HNum or raise UnknownNameException or ClassCastException.
  double? getDouble(String name, {bool checked = true}) =>
      (get(name, checked: checked) as HNum?)?.val;

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  /// String format is always "toZinc"
  @override
  String toString() => toZinc();

  /// Hash code is based on tags
  @override
  int get hashCode;

  /// Equality is tags
  @override
  bool operator ==(Object other) {
    if (other is HDict) {
      return _mapEquality.equals(this, other);
    } else {
      return false;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Encoding
//////////////////////////////////////////////////////////////////////////

  /// Returns if the given string is a legal tag name.
  ///
  /// The first char must be ASCII lower case letter. Rest of
  /// chars must be ASCII letter, digit, or underbar.
  static bool isTagName(String n) {
    if (n.isEmpty) return false;
    int first = n.codeUnitAt(0);
    if (first < a_CHAR || first > z_CHAR) return false;

    return n.codeUnits.every((int c) => c < 128 && _tagChars[c]);
  }

  static List<bool> _tagChars = () {
    List<bool> result = new List<bool>.filled(128, false);
    for (int i = a_CHAR; i <= z_CHAR; ++i) {
      result[i] = true;
    }
    for (int i = A_CHAR; i <= Z_CHAR; ++i) {
      result[i] = true;
    }
    for (int i = ZERO_CHAR; i <= NINE_CHAR; ++i) {
      result[i] = true;
    }
    result[UNDERSCORE_CHAR] = true;
    return result;
  }();

  /// Encodes value to json format
  @alwaysThrows
  @override
  String toJson() {
    throw new UnsupportedError("HDict.toJson");
  }

  /// Encodes value to zinc format
  @override
  String toZinc() {
    StringBuffer sb = new StringBuffer();
    sb..write('{')..write(_toZincMeta(this))..write('}');
    return sb.toString();
  }

  static String _toZincMeta(HDict dict) {
    StringBuffer sb = new StringBuffer();
    bool first = true;

    dict.forEach((key, val) {
      if (first) {
        first = false;
      } else {
        sb.write(' ');
      }

      sb.write(key);

      if (val != HMarker.VAL) {
        sb..write(':')
          ..write(val.toZinc());
      }
    });

    return sb.toString();
  }

  @alwaysThrows
  @override
  operator []=(String key, HVal? value) {
    throw new UnsupportedError("Cannot modify unmodifiable map");
  }

  @alwaysThrows
  @override
  void clear() {
    throw new UnsupportedError("Cannot modify unmodifiable map");
  }

  @alwaysThrows
  @override
  HVal remove(Object? key) {
    throw new UnsupportedError("Cannot modify unmodifiable map");
  }
}


final _mapEquality = MapEquality<String, HVal?>();


//////////////////////////////////////////////////////////////////////////
// MapImpl
//////////////////////////////////////////////////////////////////////////

class _MapImpl extends HDict {
  _MapImpl(this._map)
      : hashCode = _mapEquality.hash(_map),
        super._();

  @override
  int get size => _map.length;

  @override
  final int hashCode;

  @override
  HVal? operator [](Object? key) => get(key as String, checked: false);

  @override
  Iterable<String> get keys => _map.keys;

  @override
  HVal? get(String name, {bool checked = true}) {
    HVal? val = _map[name];
    if (val != null) return val;
    if (!checked) return null;
    throw new UnknownNameError(name);
  }

  final Map<String, HVal?> _map;
}
