///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hgrid;

import 'dart:collection';
import 'dart:math';

import 'package:collection/collection.dart';
import 'package:quiver/core.dart';
import 'package:meta/meta.dart';

import 'HBool.dart';
import 'HDateTime.dart';
import 'HMarker.dart';
import 'HNum.dart';
import 'HRef.dart';
import 'HStr.dart';
import 'HVal.dart';
import 'UnknownNameError.dart';
import 'io/HZincWriter.dart';
import 'string_utils.dart';

part 'HCol.dart';
part 'HDict.dart';
part 'HHisItem.dart';
part 'HRow.dart';

/// HGrid is an immutable two dimension data structure of cols and rows.
/// Use [HGridBuilder] to construct a [HGrid] instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/Grids)
class HGrid extends HVal with ListMixin<HRow> {
  /// Empty grid with one column called "empty" and zero rows.
  static final HGrid EMPTY = new HGrid(
      HDict.EMPTY,
      new List<HCol>.from(<HCol>[new HCol(0, "empty", HDict.EMPTY)]),
      <List<HVal>>[]);

  final List<HCol> cols;

  HGrid(this._meta, this.cols, List<List<HVal?>> rowList)
      : _rows = <HRow>[],
        _colsByName = new Map<String, HCol>()
  {
    for (int i = 0; i < rowList.length; ++i) {
      List<HVal?> cells = rowList[i];
      if (cols.length != cells.length) {
        throw new ArgumentError("Row cells size != cols size");
      }

      _rows.add(new HRow(this, cells));
    }

    for (int i = 0; i < cols.length; ++i) {
      HCol col = cols[i];
      String colName = col.name;
      if (_colsByName.containsKey(colName)) {
        throw new ArgumentError("Duplicate col name: $colName");
      }

      _colsByName[colName] = col;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /// Returns grid level meta.
  HDict get meta => _meta;

  UnmodifiableListView<HRow> get rows => new UnmodifiableListView(_rows.toList());

  /// Error grid have the meta.err marker tag.
  bool get isErr => meta.has("err");

  /// Returns true if number of rows is zero.
  @override
  bool get isEmpty => numRows == 0;

  /// Returns true if number of rows is more than zero.
  @override
  bool get isNotEmpty => numRows > 0;

  /// Returns number of rows
  int get numRows => _rows.length;

  /// Get a row by its zero based index.
  HRow row(int i) => _rows[i];

  /// Get number of columns.
  int get numCols => cols.length;

  @override
  int get hashCode => hash3(rows, cols, meta);

  @override
  int get length => _rows.length;

  @override
  HRow operator [](int index) => _rows[index];

  @override
  int indexOf(Object? element, [int startIndex = 0]) =>
      _rows.indexOf(element as HRow, startIndex);

  // Get a column by name.  If not found and checked if false then
  // return null, otherwise throw UnknownNameException
  HCol? colByName(String name, [bool checked = true]) {
    HCol? col = _colsByName[name];
    if (col != null) return col;
    if (checked) throw new UnknownNameError(name);
    return null;
  }

  // Get a column by its index
  HCol colByIndex(int index, [bool checked = true]) => cols[index];

//////////////////////////////////////////////////////////////////////////
// Unmodifiable interface
//////////////////////////////////////////////////////////////////////////

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void operator []=(int index, HRow value) {
    throw new UnsupportedError("Cannot modify an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  set length(int newLength) {
    throw new UnsupportedError(
        "Cannot change the length of an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void setAll(int at, Iterable<HRow> iterable) {
    throw new UnsupportedError("Cannot modify an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void add(HRow value) {
    throw new UnsupportedError("Cannot add to an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void insert(int index, HRow element) {
    throw new UnsupportedError("Cannot add to an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void insertAll(int at, Iterable<HRow> iterable) {
    throw new UnsupportedError("Cannot add to an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void addAll(Iterable<HRow> iterable) {
    throw new UnsupportedError("Cannot add to an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  bool remove(Object? element) {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void removeWhere(bool test(HRow element)) {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void retainWhere(bool test(HRow element)) {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void sort([Comparator<HRow>? compare]) {
    throw new UnsupportedError("Cannot modify an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void shuffle([Random? random]) {
    throw new UnsupportedError("Cannot modify an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void clear() {
    throw new UnsupportedError("Cannot clear an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  HRow removeAt(int index) {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  HRow removeLast() {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void setRange(int start, int end, Iterable<HRow> iterable,
      [int skipCount = 0]) {
    throw new UnsupportedError("Cannot modify an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void removeRange(int start, int end) {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void replaceRange(int start, int end, Iterable<HRow> iterable) {
    throw new UnsupportedError("Cannot remove from an unmodifiable list");
  }

  /// This operation is not supported by an unmodifiable list.
  @alwaysThrows
  @override
  void fillRange(int start, int end, [HRow? fillValue]) {
    throw new UnsupportedError("Cannot modify an unmodifiable list");
  }

//////////////////////////////////////////////////////////////////////////
// HVal
//////////////////////////////////////////////////////////////////////////

  @alwaysThrows
  @override
  String toZinc() {
    throw new UnsupportedError("HGrid.toZinc");
  }

  @alwaysThrows
  @override
  String toJson() {
    throw new UnsupportedError("HGrid.toJson");
  }

  @override
  String toString() => '[Instance of HGrid]';

  @override
  bool operator ==(Object other) {
    if (other is HGrid) {
      if (meta != other.meta) return false;
      if (cols != other.cols) return false;
      if (_rows != other._rows) return false;
      return true;
    } else {
      return false;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  /// Prints grid to [out].
  ///
  /// This is Zinc right now.
  void dump(StringBuffer out) {
    out.write(HZincWriter.gridToString(this));
  }

//////////////////////////////////////////////////////////////////////////
// Rows
//////////////////////////////////////////////////////////////////////////

  final List<HRow> _rows;

  final HDict _meta;
  final Map<String, HCol> _colsByName;
}
