///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hgrid_builder;

import 'HDictBuilder.dart';
import 'HGrid.dart';
import 'HVal.dart';

/// HGridBuilder is used to construct an immutable HGrid instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/Grids)
class HGridBuilder {
  /// Convenience to build grid from array of HDict.
  /// Any null entry will be row of all null cells.
  static HGrid dictsToGrid(HDict meta, List<HDict> dicts) {
    if (dicts.isEmpty) {
      return HGrid(meta, [HCol(0, "empty", HDict.EMPTY)], []);
    }

    HGridBuilder b = new HGridBuilder();
    b.meta.addDict(meta);

    var colsByName = collectColumns(dicts).toList();

    // if all dicts were null, handle special case
    // by creating a dummy column
    if (colsByName.isEmpty) {
      colsByName.add("empty");
    }

    colsByName.forEach((name) => b.addCol(name));

    // now map rows
    dicts.forEach((dict) {
      List<HVal?> cells = collectValues(dict, b._cols);
      b._rows.add(cells);
    });

    return b.toGrid();
  }

  /// Collects values from [dict] for given [columns].
  ///
  /// Keeps values order like it goes in [columns].
  static List<HVal?> collectValues(HDict dict, Iterable<BCol> columns) {
    return columns.map<HVal?>((col) => dict.get(col.name, checked:false)).toList();
  }

  /// Collects column names from [dicts].
  static Iterable<String> collectColumns(Iterable<HDict> dicts) {
    var colsByName = Set<String>();

    dicts.forEach((dict) {
      colsByName.addAll(dict.keys);
    });

    return colsByName;
  }

  //todo
//  /// Convenience to build an error grid from exception */
//  public static HGrid errToGrid(Throwable e)
//  {
//    // Java sucks
//    StringWriter sout = new StringWriter();
//    PrintWriter pout = new PrintWriter(sout);
//    e.printStackTrace(pout);
//    pout.flush();
//    String trace = sout.toString();
//    StringBuffer temp = new StringBuffer(trace.length());
//    for (int i=0; i<trace.length(); ++i)
//    {
//      int ch = trace.charAt(i);
//      if (ch == '\t') temp.append("  ");
//      else if (ch != '\r') temp.append((char)ch);
//    }
//    trace = temp.toString();
//
//    HGridBuilder b = new HGridBuilder();
//    b.meta().add("err")
//    .add("dis", e.toString())
//    .add("errTrace", trace);
//    b.addCol("empty");
//    return b.toGrid();
//  }

  /// Convenience to build grid from array of HHisItem.
  static HGrid hisItemsToGrid(HDict meta, List<HHisItem> items) {
    HGridBuilder b = new HGridBuilder();
    b.meta.addDict(meta);
    b.addCol("ts");
    b.addCol("val");
    for (int i = 0; i < items.length; ++i) {
      b._rows.add([items[i].ts, items[i].val]);
    }
    return b.toGrid();
  }

  HGridBuilder();

  HGridBuilder.fromGrid(HGrid grid) {
    _meta.addDict(grid.meta);
    _cols.addAll(grid.cols.map((col) => new BCol.fromHCol(col)));
    _rows.addAll(grid.map((row) => row.values.toList()));
  }

//////////////////////////////////////////////////////////////////////////
// Building
//////////////////////////////////////////////////////////////////////////

  /// Get the builder for the grid meta map.
  HDictBuilder get meta => _meta;

  /// Add new column and return builder for column metadata.
  /// Columns cannot be added after adding the first row.
  HDictBuilder addCol(String name, {HVal fill()?}) {
    if (!HDict.isTagName(name)) {
      throw new ArgumentError("Invalid column name: $name");
    }

    BCol col = new BCol(name);
    _cols.add(col);

    if (fill != null) {
      _rows.forEach((row) => row.add(fill()));
    } else {
      _rows.forEach((row) => row.add(null));
    }

    return col.meta;
  }

  /// Add new row with array of cells which correspond to column
  /// order.  Return this.
  HGridBuilder addRow(List<HVal?> cells) {
    if (_cols.length != cells.length) {
      throw new ArgumentError("Row cells size != cols size");
    }
    _rows.add(new List<HVal?>.from(cells));
    return this;
  }

  void removeCol(String name) {
    var colIndex = _cols.map((c) => c.name).toList().indexOf(name);
    if (colIndex < 0) return;

    _cols.removeAt(colIndex);
    _rows.forEach((vals) => vals.removeAt(colIndex));
  }

  HGrid toGrid() {
    // meta
    HDict meta = this._meta.toDict();

    // cols
    List<HCol> hcols = <HCol>[];
    for (int i = 0; i < _cols.length; ++i) {
      BCol bc = _cols[i];
      hcols.add(new HCol(i, bc.name, bc.meta.toDict()));
    }

    // let HGrid constructor do the rest...
    return new HGrid(meta, hcols, _rows);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final HDictBuilder _meta = new HDictBuilder();
  final List<BCol> _cols = <BCol>[];
  final List<List<HVal?>> _rows = <List<HVal?>>[];
}

//////////////////////////////////////////////////////////////////////////
// BCol
//////////////////////////////////////////////////////////////////////////

class BCol {
  BCol(this.name);

  factory BCol.fromHCol(HCol col) {
    var bcol = new BCol(col.name);
    bcol.meta.addDict(col.meta);
    return bcol;
  }

  final String name;
  final HDictBuilder meta = new HDictBuilder();
}
