///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hproj;

import 'dart:async';

import 'package:meta/meta.dart';

import 'HGrid.dart';
import 'HNum.dart';
import 'HRef.dart';
import 'HWatch.dart';
import 'UnknownRecError.dart';

/// HProj is the common interface for HClient and HServer to provide
/// access to a database tagged entity records.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel).
abstract class HProj {
//////////////////////////////////////////////////////////////////////////
// Operations
//////////////////////////////////////////////////////////////////////////

  /// Get the summary "about" information.
  Future<HDict> about();


  /// Commits a set of diffs.
  ///
  /// [operator] defines the commit operation:
  ///  /// *add*: adds new records into the database and returns a grid with the newly minted
  ///     record identifiers. As a general rule you should not have an id column in your commit
  ///     grid. However if you wish to predefine the id of the records, you can specify an id
  ///     column in the commit grid.
  ///  /// *update*: modified existing records, the records must have both an id and mod column,
  ///  /// *remove*: removes existing records, the records should have only an id and mod column,
  ///
  /// If an error occurs an error grid is returned.
  ///
  /// For add and update, the result grid contains the id and new mod. If the request grid meta
  /// defines the readReturn marker tag, then the response contains the full tag definitions of
  /// the new/updated records.
  ///
  /// Commit access requires admin permission for the project.
  ///
  /// Note you can also commit by evaluating axon expressions which use the commit function. But
  /// with large datasets committing directly through REST API can be more efficient.
  Future<HGrid> commit(Iterable<HDict> diffs, CommitOperator operator, {bool force = false});


//////////////////////////////////////////////////////////////////////////
// Read by id
//////////////////////////////////////////////////////////////////////////

  /// Calls "read" to lookup an entity record by it's unique identifier.
  ///
  /// If not found then return null or throw an UnknownRecException based
  /// on checked.
  Future<HDict?> readById(HRef id, {bool checked = true}) {
    return onReadById(id).then((HDict? rec) {
      if (rec != null) return rec;
      if (checked) throw new UnknownRecError(id: id);
      return null;
    });
  }

  /// Reads a list of entity records by their unique identifier.
  ///
  /// Returns a grid where each row of the grid maps to the respective
  /// id array (indexes line up).  If checked is true and any one of the
  /// ids cannot be resolved then raise UnknownRecException for first id
  /// not resolved.  If checked is false, then each id not found has a
  /// row where every cell is null.
  Future<HGrid> readByIds(List<HRef> ids, {bool checked = true}) {
    return onReadByIds(ids).then((HGrid grid) {
      if (checked) {
        for (int i = 0; i < grid.numRows; ++i) {
          if (grid.row(i).missing("id")) throw new UnknownRecError(id: ids[i]);
        }
      }
      return grid;
    });
  }

  /// Subclass hook for readById, return null if not found.
  @protected
  Future<HDict?> onReadById(HRef id);

  /// Subclass hook for readByIds, return rows with nulls cells
  /// for each id not found.
  @protected
  Future<HGrid> onReadByIds(List<HRef> ids);

//////////////////////////////////////////////////////////////////////////
// Read by filter
//////////////////////////////////////////////////////////////////////////

  /// Query one entity record that matches the given filter.
  ///
  /// If there is more than one record, then it is undefined which one is
  /// returned.  If there are no matches than return null or raise
  /// UnknownRecException based on checked flag.
  Future<HDict?> read(String filter, {bool checked = true}) {
    return readAll(filter, limit: 1).then((HGrid grid) {
      if (grid.isNotEmpty) return grid.row(0);

      if (checked) throw new UnknownRecError(msg: filter);

      return null;
    });
  }

  /// Calls "read" to query every entity record that matches given filter.
  ///
  /// Clips number of results by "limit" parameter. Default is 2147483647 - max 32 bit int value.
  Future<HGrid> readAll(String filter, {int limit = 2147483647}) =>
      onReadAll(filter, limit);

  /// Subclass hook for read and readAll.
  @protected
  Future<HGrid> onReadAll(String filter, int limit);

//////////////////////////////////////////////////////////////////////////
// Watches
//////////////////////////////////////////////////////////////////////////

  /// Creates a new watch with an empty subscriber list.
  ///
  /// The dis string is a debug string to keep track of who created the watch.
  /// Pass the desired lease time or null to use default.
  HWatch watchOpen(String dis, HNum lease);

  /// List the open watches.
  List<HWatch> watches();

  /// Looks up a watch by its unique identifier.
  ///
  /// If not found then raise UnknownWatchErr or return null based on checked flag.
  HWatch watch(String id, {bool checked = true});

//////////////////////////////////////////////////////////////////////////
// Historian
//////////////////////////////////////////////////////////////////////////

//  /// Read history time-series data for given record and time range. The
//  /// items returned are exclusive of start time and inclusive of end time.
//  /// Raise exception if id does not map to a record with the required tags
//  /// "his" or "tz".  The range may be either a String or a HDateTimeRange.
//  /// If HTimeDateRange is passed then must match the timezone configured on
//  /// the history record.  Otherwise if a String is passed, it is resolved
//  /// relative to the history record's timezone.
//  Future<HGrid> hisRead(HRef id, Object range);
//
//  /// Write a set of history time-series data to the given point record.
//  /// The record must already be defined and must be properly tagged as
//  /// a historized point.  The timestamp timezone must exactly match the
//  /// point's configured "tz" tag.  If duplicate or out-of-order items are
//  /// inserted then they must be gracefully merged.
//  void hisWrite (HRef id, List<HHisItem> items);
}


enum CommitOperator {
  add,
  update,
  remove
}

String commitToString(CommitOperator operator) {
  switch (operator) {
    case CommitOperator.add: return 'add';
    case CommitOperator.update: return 'update';
    case CommitOperator.remove: return 'remove';
    default:
      throw new ArgumentError('Operator should be valid CommitOperator value.');
  }
}