///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.href;

import 'HVal.dart';
import 'HStr.dart';
import 'string_utils.dart';

/// HRef wraps a string reference identifier and optional display name.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HRef extends HVal {
  factory HRef(String val, [String? dis]) {
    if (!isId(val)) {
      throw new FormatException("Invalid id val: \"$val\"");
    }
    return new HRef._(val, dis);
  }

  HRef._(this.val, this.dis);

  /// String identifier for reference.
  final String val;

  /// Display name for reference or null.
  final String? dis;

  /// Hash code is based on val field only.
  @override
  int get hashCode => val.hashCode;

  /// Equals is based on val field only.
  @override
  bool operator ==(Object that) => that is HRef && this.val == that.val;

  /// Returns display string which is dis field if non-null, val field otherwise.
  String getDis() => (dis != null) ? dis! : val;

  /// Returns the val string.
  @override
  String toString() => val;

  /// Encodes as "@id".
  String toCode() => "@$val";

  /// Encodes as "r:<id> [dis]".
  @override
  String toJson() {
    StringBuffer s = new StringBuffer();
    s..write("r:")..write(val);
    if (dis != null) {
      s..write(' ')..write(dis);
    }
    return s.toString();
  }

  /// Encodes as "@id <dis>"
  @override
  String toZinc() {
    StringBuffer s = new StringBuffer();
    s.write('@');
    s.write(val);

    if (dis != null) {
      s.write(' ');
      HStr.encodeForZinc(s, dis!);
    }

    return s.toString();
  }

  /// Returns if the given string is a valid id for a reference.
  static bool isId(String id) {
    if (id.isEmpty) return false;

    return id.codeUnits.every((int c) {
      return isIdChar(c);
    });
  }

  /// Is the given character valid in the identifier part.
  static bool isIdChar(int ch) =>
      ch >= 0 && ch < _idChars.length && _idChars[ch];

  static final List<bool> _idChars = () {
    List<bool> result = List<bool>.filled(127, false);
    for (int i = a_CHAR; i <= z_CHAR; ++i) {
      result[i] = true;
    }
    for (int i = A_CHAR; i <= Z_CHAR; ++i) {
      result[i] = true;
    }
    for (int i = ZERO_CHAR; i <= NINE_CHAR; ++i) {
      result[i] = true;
    }
    result[UNDERSCORE_CHAR] = true;
    result[COLON_CHAR] = true;
    result[MINUS_CHAR] = true;
    result[PERIOD_CHAR] = true;
    result[TILDE_CHAR] = true;

    return result;
  }();
}
