///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hwatch;

import 'dart:async';

import 'HGrid.dart';
import 'HNum.dart';
import 'HRef.dart';

/// HWatch models a subscription to a list of entity records.
///
/// Use [HProj.watchOpen] to create a new watch.
///
/// See [Project Haystack](http://project-haystack.org/doc/Rest#watches)
abstract class HWatch {
  /// Unique watch identifier within a project database.
  ///
  /// The id may not be assigned until after the first call
  /// to "sub", in which case return null.
  String get id;

  /// Debug display string used during "HProj.watchOpen"
  String get dis;

  /// Lease period or null if watch has not been opened yet.
  HNum lease();

  /// Add a list of records to the subscription list and return their
  /// current representation.
  ///
  /// If checked is true and any one of the
  /// ids cannot be resolved then raise UnknownRecException for first id
  /// not resolved.  If checked is false, then each id not found has a
  /// row where every cell is null.
  ///
  /// The HGrid that is returned must contain metadata entries
  /// for 'watchId' and 'lease'.
  Future<HGrid> sub(List<HRef> ids, {bool checked = true});

  /// Remove a list of records from watch.
  ///
  /// Silently ignore any invalid ids.
  Future<HGrid> unsub(List<HRef> ids);

  /// Poll for any changes to the subscribed records.
  Future<HGrid> pollChanges();

  /// Poll all the subscribed records even if there have been no changes.
  Future<HGrid> pollRefresh();

  /// Close the watch and free up any state resources.
  Future close();

  /// Returns whether this watch is currently open.
  bool get isOpen;
}
