import 'package:haystack/haystack.dart';
import 'package:haystack/src/HGrid.dart';
import 'package:haystack/src/io/HGridWriter.dart';
import 'package:haystack/src/string_utils.dart';

/// Used to write grids in the Trio format.
///
/// See [Project Haystack](https://project-haystack.org/doc/Trio)
class HTrioWriter implements HGridWriter {


//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  HTrioWriter(this._out);


//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

  // Write the grid rows (no support for meta).
  @override
  void writeGrid(HGrid grid) {
    grid.forEach((HRow rec) => writeRec(rec));
  }


//  Write separator and record.
  void writeRec(HDict rec) {
    if (_needSep) _out.writeln("---");
    else _needSep = true;

    // get names in nice order
    List<String> names = rec.keys.toList();
    if (names.contains("name")) {
      names.remove("name");
      names.insert(0, "name");
    }

    if (names.contains("dis")) {
      int disIndex = names.contains("name") ? 1 : 0;
      names.remove("dis");
      names.insert(disIndex, "dis");
    }

    if (names.contains("src")) {
      names.remove("src");
      names.add("src");
    }

    names.forEach((String n) {
      HVal? tagVal = rec[n];
      if (tagVal == null) return;

      if (tagVal == HMarker.VAL) {
        _out.writeln(n);
        return;
      }

      tagVal = normVal(tagVal) as HVal?;
      _out.write(n);
      _out.write(":");

      if (tagVal is! HStr) {
        bool isCollection = tagVal is HDict || tagVal is Iterable; //fixme: not sure if it is correct check
        if (isCollection) _writeCollection(tagVal);
        else if (tagVal is HBool) _out.writeln(tagVal);
        else _out.writeln(tagVal!.toZinc());
        return;
      }

      String s = tagVal.toString();
      if (!s.contains("\n")) {
        if (useQuotes(s)) _out.writeln("\"$s\"");
        else _out.writeln(s);
      } else {
        _out.writeln();
        s.split("\n").forEach((String line) {
          _out.write("  ");
          _out.writeln(line);
        });
      }
    });
  }

  void _writeCollection(Object? val) {
    // Write anything with nested grids as indented zinc.
    if (_requiresNestedZinc(val)) return _writeNestedZinc(val);

    HZincWriter zincWriter = HZincWriter(_out);
    zincWriter.writeVal(val as HVal?);
    _out.writeln();
  }

  bool _requiresNestedZinc(Object? val) {
    // Grids are multi-line so must nest.
    if (val is HGrid) return true;

    // List requires checking each list value.
    if (val is List) return val
        .any(((Object val) => _requiresNestedZinc(val)) as bool Function(dynamic));

    if (val is HDict) {
      var r;
      val.forEach((String name, HVal tagVal) {
        r = _requiresNestedZinc(tagVal) ? "req" : null;
      });

      return r != null;
    }

    return false;
  }

  void _writeNestedZinc(Object? val) {
    StringBuffer stringBuffer = StringBuffer();
    HZincWriter zincWriter = HZincWriter(stringBuffer);

    if (val is HGrid) {
      zincWriter.writeGrid(val);
    } else {
      zincWriter.writeVal(val as HVal?);
    }

    _out.writeln("Zinc:");
    stringBuffer.toString()
        .split("\n")
        .forEach((String line) {
          _out.write("  ");
          _out.writeln(line);
        });
  }

  static bool useQuotes(String s) {
    if (s.isEmpty) return true;
    if (s.codeUnitAt(0) > 127) return false;
    if (isAlpha(s[0])) return true;
    if (quotedKeyword.contains(s)) return true;

    for (int i = 0; i < s.length; i++) {
      if (!_requireQuoteChar(s.codeUnitAt(i))) return true;
    }

    return false;
  }

  static bool _requireQuoteChar(int ch) {
    if (ch > 127) return false;
    return quoteChars[ch];
  }

  // Return if every char is an ASCII letter.
  static bool isAlpha(String s) {
    bool result = true;

    for(int i = 0; i < s.length; i ++) {
      int codeUnit = s.codeUnitAt(i);

      if (!isAlphaInt(codeUnit)) {
        result = false;
        break;
      }
    }

    return result;
  }

  static bool isAlphaInt(int i) {
    bool isLowerCaseLetter = i >= a_CHAR && i <= z_CHAR;
    bool isUpperCaseLetter = i >= A_CHAR && i <= Z_CHAR;
    return isLowerCaseLetter || isUpperCaseLetter;
  }


//  Write the list of records.
  void writeAllRecs(List<HDict> recs) {
    recs.forEach((HDict rec) => writeRec(rec));
  }

//  Hook to normalize value
  Object normVal (Object val) => val;

  static List<bool> _defineQuotedChars () {
    List<bool> result = List.filled(127, false);

    for (int i = a_CHAR; i <= z_CHAR; i ++) result[i] = true;
    for (int i = A_CHAR; i <= Z_CHAR; i ++) result[i] = true;
    for (int i = ZERO_CHAR; i <= NINE_CHAR; i ++) result[i] = true;
    result[SPACE_CHAR] = true;
    result[MINUS_CHAR] = true;
    result[UNDERSCORE_CHAR] = true;

    return result;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////
  static const List<String> quotedKeyword = [
    "true",
    "false",
    "INF",
    "NA",
    "NaN",
    "R",
    "Zinc"
  ];

  static List<bool> quoteChars = _defineQuotedChars();
  final StringBuffer _out;
  bool _needSep = false;
}