///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

import 'package:haystack/src/HDate.dart';
import 'package:haystack/src/HDateTime.dart';
import 'package:haystack/src/HNum.dart';
import 'package:haystack/src/HRef.dart';
import 'package:haystack/src/HStr.dart';
import 'package:haystack/src/HTime.dart';
import 'package:haystack/src/HUri.dart';
import 'package:haystack/src/ParseError.dart';
import 'package:haystack/src/io/trio_token.dart';
import 'package:haystack/src/string_utils.dart';
import 'package:unicode/unicode.dart' as unicode;

import 'trio_token.dart';

/// Stream based tokenizer for Haystack formats such as Zinc and Filters.
class TrioTokenizer {
//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////
  static const int eof = -1;
  TrioToken? tok; //  _current token
//  Object val; //token literal or identifier
  int line = 1; //  _current line number

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  TrioTokenizer(Iterable<int> inStream)
      : this._inStream = inStream.iterator {
        this.tok = TrioToken.eof;
    _consume();
    _consume();
  }

  bool close() {
//    try {
//      //_inStream.close();
    return true;
//    } catch (IOException e) {
//      return false;
//    }
  }

//////////////////////////////////////////////////////////////////////////
// Tokenizing
//////////////////////////////////////////////////////////////////////////

  // todo: check if it throws errors in any cases.
  // todo: add docs and check usages to add try-catch.
  void next() {
    // reset
//    val = null;

    // Skip comments.
    // Trio supports only single-line comments.
    while (true) {
      bool isCommentStarts = _cur == SLASH_CHAR && _peek == SLASH_CHAR;
      if (isCommentStarts) {
        _skipCommentsSL();
        continue;
      }

      break;
    }

    // Handle newlines and entities delimiter.
    if (_cur == NL_CHAR || _cur == CR_CHAR) {
      if (_cur == CR_CHAR && _peek == NL_CHAR) {
        _consume('\r');
      }
      _consume();
      ++line;

      bool isDelimiterStarts = _cur == MINUS_CHAR;
      if (isDelimiterStarts) tok = _handleEntriesDelimiter();
      else tok = TrioToken.nl;
      return;
    }

    // Handle various starting chars.
    // todo: handle coordinates (C(37.5458,-77.4491))
    if (_isIdStart(_cur)) tok = _handleId();
    if (_isIdStart(_cur)) tok = _handleId();
    else if (_cur == DOG_CHAR) tok = _handleRef();
    else if (_cur == DOUBLE_QUOTE_CHAR) tok = _handleQuotedStr();
    else if (_isDigit(_cur)) tok = _handleNum();
    else if (_cur == MINUS_CHAR && _isDigit(_peek)) tok = _handleNum();
    else if (_cur == BACKQUOTE_CHAR) tok = _handleUri();
    else tok = _handleSymbol();
  }


//////////////////////////////////////////////////////////////////////////
// Token Productions
//////////////////////////////////////////////////////////////////////////

  TrioToken _handleEntriesDelimiter() {
    _skipDelimiter();
    return TrioToken.trioEntitiesDelimiter;
  }

  void _skipDelimiter() {
    while (!_isEntitiesDelimiterEnds()) _consume();
    _consume(); // skip last minus of the delimiter.
  }

  // In Trio entities are separated by lines beginning with "-".
  // Line can contain any number of "-".
  bool _isEntitiesDelimiterEnds() {
    bool nextIsNl = _peek == NL_CHAR || _peek == CR_CHAR || _peek == eof;
    return _cur == MINUS_CHAR && nextIsNl;
  }

  // todo: in zinc Strings are always quoted. So there is difference between String and identifier.
  // todo: but in Trio strings can be unquoted in some cases.
  // todo: need to destinguish cases of identifier and string
  // todo: maybe we can do it on reader level.
  TrioToken _handleId() {
    StringBuffer s = StringBuffer();

    while (_isIdPart(_cur)) {
      s.writeCharCode(_cur);
      _consume();
    }

    String idString = s.toString();
    return TrioToken.data(TokenType.identifier, "identifier", idString);
  }

  /// Returns if cur char is allowed start of identifier.
  ///
  /// Only lower case latin letters are allowed.
  static bool _isIdStart(int _cur) {
    return a_CHAR <= _cur && _cur <= z_CHAR;
  }

//  /// Returns if cur char is allowed start of word.
//  ///
//  /// Only lower case latin letters are allowed.
//  static bool _isIdStart(int _cur) {
//    return a_CHAR <= _cur && _cur <= z_CHAR;
//  }

  /// Returns if cur char is allowed part of identifier (not a start).
  ///
  /// Allowed symbols:
  /// - lower case latin letter,
  /// - upper case latin letter,
  /// - digit,
  /// - underscore
  static bool _isIdPart(int _cur) {
    bool validLetter =  a_CHAR <= _cur && _cur <= z_CHAR || A_CHAR <= _cur && _cur <= Z_CHAR;
    if (validLetter) return true;
    if (_isDigit(_cur)) return true;
    if (_cur == UNDERSCORE_CHAR) return true;
    return false;
  }

  static bool _isDigit(int c) => (c >= ZERO_CHAR && c <= NINE_CHAR);
  static bool _isLetter(int c) => (c != eof && isLetter(c));

  /// Handles tokens which start with digit.
  /// It can be:
  /// - numbers
  /// - date
  /// - time
  /// - dateTime
  TrioToken _handleNum() {
    TrioToken result;

    // Parse hex number (no unit allowed).
    bool isHex = (_cur == ZERO_CHAR && _peek == x_CHAR);
    if (isHex) result = _handleHexNum();

    //  _consume all things that might be part of this number token
    StringBuffer s = StringBuffer()..writeCharCode(_cur);
    _consume();
    int colons = 0;
    int dashes = 0;
    int unitIndex = 0;
    bool exp = false;

    while (true) {
      if (!_isDigit(_cur)) {
        if (exp && (_cur == PLUS_CHAR || _cur == MINUS_CHAR)) {
        } else if (_cur == MINUS_CHAR) {
          ++dashes;
        } else if (_cur == COLON_CHAR && _isDigit(_peek)) {
          ++colons;
        } else if ((exp || colons >= 1) && _cur == PLUS_CHAR) {
        } else if (_cur == PERIOD_CHAR) {
          if (!_isDigit(_peek)) break;
        } else if ((_cur == e_CHAR || _cur == E_CHAR) &&
            (_peek == MINUS_CHAR || _peek == PLUS_CHAR || _isDigit(_peek))) {
          exp = true;
        } else if (_isLetter(_cur) ||
            _cur == PERCENT_CHAR ||
            _cur == DOLLAR_CHAR ||
            _cur == SLASH_CHAR ||
            _cur > 128) {
          if (unitIndex == 0) {
            unitIndex = s.length;
          }
        } else if (_cur == UNDERSCORE_CHAR) {
          if (unitIndex == 0 && _isDigit(_peek)) {
            _consume();
            continue;
          } else {
            if (unitIndex == 0) {
              unitIndex = s.length;
            }
          }
        } else {
          break;
        }
      }

      s.writeCharCode(_cur);
      _consume();
    }

    if (dashes == 2 && colons == 0) result = _handleDate(s.toString());
    else if (dashes == 0 && colons >= 1) result = _handleTime(s, colons == 1);
    else if (dashes >= 2) result = _handleDateTime(s);
    else result = _handleNumber(s.toString(), unitIndex);

    return result;
  }

  static bool _isHex(int c) {
    if (c == eof) return false;
    c = String.fromCharCode(c).toLowerCase().codeUnitAt(0);
    if (a_CHAR <= c && c <= f_CHAR) return true;
    if (_isDigit(c)) return true;
    return false;
  }

  TrioToken _handleHexNum () {
    _consume('0');
    _consume('x');
    StringBuffer s = StringBuffer();

    while (true) {
      if (_isHex(_cur)) {
        s.writeCharCode(_cur);
        _consume();
        continue;
      }
      if (_cur == UNDERSCORE_CHAR) {
        _consume();
        continue;
      }
      break;
    }

    HNum resultVal = HNum(int.parse(s.toString(), radix: 16));
    TrioToken result = TrioToken.data(TokenType.num, "Num", resultVal);

    return result;
  }

  TrioToken _handleDate(String s) {
    try {
      HDate dateVal = HDate.fromString(s);
      return TrioToken.data(TokenType.date, "DateTime", dateVal);

    } catch (e) {
      throw _err(e.toString());
    }
  }

  /// We don't require hour to be two digits and we don't require seconds.
  TrioToken _handleTime(StringBuffer s, bool addSeconds) {
    try {
      var str = s.toString();

      if (str[1] == ":") {
        str = "0$str";
      }
      if (addSeconds) {
        str = "$str:00";
      }

      HTime timeVal = HTime.fromString(str);
      return TrioToken.data(TokenType.time, "Time", timeVal);

    } catch (e) {
      throw _err(e.toString());
    }
  }

  TrioToken _handleDateTime(StringBuffer s) {
    // xxx timezone
    if (_cur != SPACE_CHAR || !unicode.isUppercaseLetter(_peek)) {
      if (s.toString()[s.length - 1] == 'Z') {
        s.write(" UTC");
      } else {
        throw _err("Expecting timezone");
      }
    } else {
      _consume();
      s.write(' ');
      while (_isIdPart(_cur)) {
        s.writeCharCode(_cur);
        _consume();
      }

      // handle GMT+xx or GMT-xx
      if ((_cur == PLUS_CHAR || _cur == MINUS_CHAR) &&
          s.toString().endsWith("GMT")) {
        s.writeCharCode(_cur);
        _consume();
        while (_isDigit(_cur)) {
          s.writeCharCode(_cur);
          _consume();
        }
      }
    }

    try {
      HDateTime dateTimeVal = HDateTime.fromString(s.toString());
      return TrioToken.data(TokenType.dateTime, "DateTime", dateTimeVal);

    } catch (e) {
      throw _err(e.toString());
    }
  }

  /// Handles number from string.
  ///
  /// If number doesn't have a unit [unitIndex] should be 0.
  TrioToken _handleNumber(String s, int unitIndex) {
    try {
      String numStr;
      String? unitStr;

      bool hasUnit = unitIndex != 0;
      if (hasUnit) {
        numStr = s.substring(0, unitIndex);
        unitStr = s.substring(unitIndex);
      } else {
        numStr = s;
      }

      HNum numVal = HNum(num.parse(numStr), unitStr);
      return TrioToken.data(TokenType.num, "Num", numVal);

    } catch (e) {
      throw _err("Invalid Number literal: $s");
    }
  }

  //todo: rework for trio format
  TrioToken _handleQuotedStr() {
    _consume('"');
    StringBuffer s = StringBuffer();

    while (true) {
      if (_cur == eof) throw _err("Unexpected end of str");
      if (_cur == DOUBLE_QUOTE_CHAR) {
        _consume('"');
        break;
      }
      if (_cur == BACKSLASH_CHAR) {
        s.writeCharCode(_escape());
        continue;
      }
      s.writeCharCode(_cur);
      _consume();
    }

    HStr strVal = HStr(s.toString());
    return TrioToken.data(TokenType.str, "Str", strVal);
  }


  TrioToken _handleRef() {
    _consume('@');
    StringBuffer s = StringBuffer();

    // Parsing val part of id.
    while (true) {
      if (HRef.isIdChar(_cur)) {
        s.writeCharCode(_cur);
        _consume();
      } else {
        break;
      }
    }

    // Skip non-meaningful whitespaces.
    while (true) {
      // treat space, tab, non-breaking space as whitespace
      if (_cur == SPACE_CHAR || _cur == TAB_CHAR || _cur == 0xa0) {
        _consume();
        continue;
      }

      break;
    }

    // Parsing dis part of id if exists.
    TrioToken? dis;
    if (_cur == DOUBLE_QUOTE_CHAR) dis = _handleQuotedStr();
    assert (dis == null || dis.value is String);

    HRef resultRef = HRef(s.toString(), dis!.value as String?);
    return TrioToken.data(TokenType.ref, "Ref", resultRef);
  }

  TrioToken _handleUri() {
    _consume('`');
    StringBuffer s = StringBuffer();

    while (true) {
      if (_cur == BACKQUOTE_CHAR) {
        _consume('`');
        break;
      }
      if (_cur == eof || _cur == NL_CHAR) throw _err("Unexpected end of uri");
      if (_cur == BACKSLASH_CHAR) {
        /// TODO or convert all chars to int ASC codes in below cases:
        switch (_peek) {
          case COLON_CHAR:
          case SLASH_CHAR:
          case QUESTION_MARK_CHAR:
          case NUMBER_SIGN_CHAR:
          case OPEN_BRACKET_CHAR:
          case CLOSE_BRACKET_CHAR:
          case DOG_CHAR:
          case BACKSLASH_CHAR:
          case AMPERSAND_CHAR:
          case EQUALS_SIGN_CHAR:
          case SEMICOLON_CHAR:
            s.writeCharCode(_cur);
            s.writeCharCode(_peek);
            _consume();
            _consume();
            break;
          default:
            s.writeCharCode(_escape());
        }
      } else {
        s.writeCharCode(_cur);
        _consume();
      }
    }

    HUri uriVal = HUri(s.toString());
    return TrioToken.data(TokenType.uri, "Uri", uriVal);
  }

  int _escape() {
    _consume('\\');
    switch (_cur) {
      case b_CHAR:
        _consume();
        return BS_CHAR; // \b
      case f_CHAR:
        _consume();
        return FF_CHAR; // \f
      case n_CHAR:
        _consume();
        return NL_CHAR;
      case r_CHAR:
        _consume();
        return CR_CHAR;
      case t_CHAR:
        _consume();
        return TAB_CHAR;
      case DOUBLE_QUOTE_CHAR:
        _consume();
        return DOUBLE_QUOTE_CHAR;
      case DOLLAR_CHAR:
        _consume();
        return DOLLAR_CHAR;
      case QUOTE_CHAR:
        _consume();
        return '\''.codeUnitAt(0);
      case BACKQUOTE_CHAR:
        _consume();
        return BACKQUOTE_CHAR;
      case BACKSLASH_CHAR:
        _consume();
        return BACKSLASH_CHAR;
    }

    // check for uxxxx
    StringBuffer esc = new StringBuffer();
    if (_cur == u_CHAR) {
      _consume('u');
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();

      try {
        return int.parse(esc.toString(), radix: 16);
      } on FormatException {
        throw new ParseError("Invalid unicode escape: $esc");
      }
    }

    throw _err("Invalid escape sequence: " + new String.fromCharCode(_cur));
  }

  TrioToken _handleSymbol() {
    print("_handleSymbol: ${_cur}");
    TrioToken result;

    int c = _cur;
    _consume();

    switch (c) {
      case COMMA_CHAR: //,
        result = TrioToken.comma;
        break;

      case COLON_CHAR: //:
        result = TrioToken.colon;
        break;

      case SEMICOLON_CHAR: // ;
        result = TrioToken.semicolon;
        break;

      case OPEN_BRACKET_CHAR: // [
        result = TrioToken.lbracket;
        break;

      case CLOSE_BRACKET_CHAR: //  ]
        result =  TrioToken.rbracket;
        break;

      case OPEN_BRACE_CHAR: // {
        result = TrioToken.lbrace;
        break;

      case CLOSE_BRACE_CHAR: // }
        result = TrioToken.rbrace;
        break;

      case OPEN_PARENTHESE_CHAR: // (
        result = TrioToken.lparen;
        break;

      case CLOSE_PARENTHESE_CHAR: // )
        result = TrioToken.rparen;
        break;

      case LESS_CHAR: // <
        if (_cur == LESS_CHAR) {
          _consume('<');
          result = TrioToken.lt2;
          break;
        }
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          result = TrioToken.ltEq;
          break;
        }
        result = TrioToken.lt;
        break;

      case GREATER_CHAR: // >
        if (_cur == GREATER_CHAR) {
          _consume('>');
          result = TrioToken.gt2;
          break;
        }
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          result = TrioToken.gtEq;
          break;
        }
        result = TrioToken.gt;
        break;

      case MINUS_CHAR: // -
        if (_cur == GREATER_CHAR) {
          _consume('>');
          result = TrioToken.arrow;
          break;
        }
        result = TrioToken.minus;
        break;

      case EQUALS_SIGN_CHAR: // =
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          result = TrioToken.eq;
          break;
        }
        result = TrioToken.assign;
        break;

      case EXCLAMATION_CHAR: // !
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          result = TrioToken.notEq;
          break;
        }
        result = TrioToken.bang;
        break;

      case SLASH_CHAR: // /
        result = TrioToken.slash;
        break;

      case TAB_CHAR:
      case SPACE_CHAR:
        result = TrioToken.space;
        break;

      case eof:
        result = TrioToken.eof;
        break;

      default:
        String errMsg = "Unexpected symbol: '${String.fromCharCode(c)}' (0x${c.toRadixString(16)})";
        throw _err(errMsg);
    }

    print("handleSymbol returns: $result");
    return result;
  }

//////////////////////////////////////////////////////////////////////////
// Comments
//////////////////////////////////////////////////////////////////////////

  void _skipCommentsSL() {
    _consume('/');
    _consume('/');
    while (true) {
      if (_cur == NL_CHAR || _cur == eof) break;
      _consume();
    }
  }

  void _skipCommentsML() {
    _consume('/');
    _consume('*');
    int depth = 1;
    while (true) {
      if (_cur == ASTERISK_CHAR && _peek == SLASH_CHAR) {
        _consume('*');
        _consume('/');
        depth--;
        if (depth <= 0) break;
      }
      if (_cur == SLASH_CHAR && _peek == ASTERISK_CHAR) {
        _consume('/');
        _consume('*');
        depth++;
        continue;
      }
      if (_cur == NL_CHAR) {
        ++line;
      }
      if (_cur == eof) throw _err("Multi-line comment not closed");
      _consume();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Error Handling
//////////////////////////////////////////////////////////////////////////

  ParseError _err(String msg) => ParseError("$msg [line $line]");

//////////////////////////////////////////////////////////////////////////
// Char
//////////////////////////////////////////////////////////////////////////

  void _consume([String? expected]) {
    if (expected != null && _cur != expected.codeUnitAt(0)) {
      throw _err("Expected $expected");
    }

    try {
      _cur = _peek;
      _inStream.moveNext();
      _peek = _inStream.current;
    } catch (e) {
      _cur = eof;
      _peek = eof;
    }
  }

  final Iterator<int> _inStream; // underlying stream
  int _cur = eof; //  _current char
  int _peek = eof; // next char
}
