///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.grid_test;

import 'package:haystack/haystack.dart';
import 'package:test/test.dart';

void main () {
  test('Test HVal', () {
    HGridBuilder b = new HGridBuilder();
    b.addCol("id");
    b.addCol("dis");
    b.addCol("area");
    b.addRow([new HRef("a"), new HStr("Alpha"), new HNum(1200)]);
    b.addRow([new HRef("b"), new HStr("Beta"), null ]);

    HGrid g = b.toGrid();

    expect(g, isNotNull);
    expect(g, equals(g));
    expect(g == g, isTrue, reason: "HGrid is not equal to itself");
  });

  test('Test empty', () {
    HGrid g = new HGridBuilder().toGrid();
    expect(g.meta, equals(HDict.EMPTY));
    expect(g.numRows, equals(0));
    expect(g.isEmpty, equals(true));
    expect(g.colByName("foo", false), null);
    expect(() => g.colByName("foo"), throwsA(const TypeMatcher<UnknownNameError>()));
  });

  group('Test no rows', () {
    HGridBuilder b = new HGridBuilder();
    b.meta.add("dis", "Title");
    b.addCol("a").add("dis", "Alpha");
    b.addCol("b");
    HGrid g = b.toGrid();

    test('meta', () {
      expect(g.meta.size, equals(1));
      expect(g.meta.get("dis"), equals(new HStr("Title")));
    });

    test('cols', () {
      HCol c;
      expect(g.numCols, equals(2));
      c = verifyCol(g, 0, "a");
      expect(c.dis, equals("Alpha"));
      expect(c.meta.size, equals(1));
      expect(c.meta.get("dis"), equals(new HStr("Alpha")));
    });

    test('rows', () {
      expect(g.numRows, equals(0));
      expect(g.isEmpty, equals(true));
    });

    test('iterator', () {
      verifyGridIterator(g);
    });

  });

  group('Test simple', () {
    HGridBuilder b = new HGridBuilder();
    b.addCol("id");
    b.addCol("dis");
    b.addCol("area");
    b.addRow([new HRef("a"), new HStr("Alpha"), new HNum(1200)]);
    b.addRow([new HRef("b"), new HStr("Beta"), null ]);

    HGrid g = b.toGrid();

    test('meta', () {
      expect(g.meta.size, equals(0));
    });

    test('cols', () {
      expect(g.numCols, equals(3));
      verifyCol(g, 0, "id");
      verifyCol(g, 1, "dis");
      verifyCol(g, 2, "area");
    });

    test('rows', () {
      expect(g.numRows, equals(2));
      expect(g.isEmpty, equals(false));

      HRow r;
      r = g.row(0);
      expect(r["id"],   equals(new HRef("a")));
      expect(r.get("dis"),  equals(new HStr("Alpha")));
      expect(r.get("area"), equals(new HNum(1200)));

      r = g.row(1);
      expect(r.get("id"),   equals(new HRef("b")));
      expect(r.get("dis"),  equals(new HStr("Beta")));

      expect(r.get("area", checked:false), equals(null));
      expect(() => r.get("area"), throwsA(const TypeMatcher<UnknownNameError>()));

      expect(r.get("fooBar", checked:false), equals(null));
      expect(() => r.get("fooBar"), throwsA(const TypeMatcher<UnknownNameError>()));
    });

    test('HRow.iterator no-nulls', () {
      final row = g.row(0);

      verifyRowByIndex(row, 0, "id",   new HRef("a"));
      verifyRowByIndex(row, 1, "dis",  new HStr("Alpha"));
      verifyRowByIndex(row, 2, "area", new HNum(1200));

      expect(row.length, equals(3));
    });

    test('HRow.iterator with nulls', () {
      final row = g.row(1);

      verifyRowByIndex(row, 0, "id",  new HRef("b"));
      verifyRowByIndex(row, 1, "dis", new HStr("Beta"));

      expect(row.length, equals(3));

      //expect(g.row(1), { "id": new HRef("b"), "dis": new HStr("Beta") });
    });

    test('iterator', () {
      verifyGridIterator(g);
    });

    test('iterable grid', () {
      expect( g.rows.map( (HRow row) => row.size ), [3, 3] );
      expect( () => g.rows[0] = new HRow(g, []), throwsUnsupportedError );
    });
  });

}

HCol verifyCol(HGrid g, int i, String n) {
  HCol col = g.colByIndex(i);
  expect(identical(g.colByIndex(i), g.colByName(n)), equals(true));
  expect(col.name, equals(n));
  return col;
}

void verifyRowByIndex(HRow row, int index, String name, HVal val) {
  expect(row.length, greaterThan(index));

  expect(row.keys, contains(name));

  expect(row, containsPair(name, val));
}

void verifyGridIterator(HGrid g) {
  Iterator<HRow> it = g.iterator;
  int c = 0;

  while (c < g.numRows) {
    expect(it.moveNext(), equals(true));
    expect(identical(it.current, g.row(c++)), equals(true));
  }

  expect(it.moveNext(), equals(false));
  expect(c, equals(g.numRows));
}